from setuptools import setup

setup(
    name='protoyang',
    version='0.0.2',
    packages=['protoyang', 'protoyang.hpproto'],
    url='https://github.com/HuaweiDatacomm/protoyang',
    license='Apache License, Version 2.0',
    author='feng chong',
    author_email='frank.fengchong@huawei.com',
    description='a tool generate .proto files according yang modules'
)
